/*
    Copyright (C) 2008, 2009 William Hart
    Copyright (C) 2010 Fredrik Johansson

    This file is part of FLINT.

    FLINT is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#include "fmpz_factor.h"

void
fmpz_factor_init(fmpz_factor_t factor)
{
    factor->sign  = 1;
    factor->p     = NULL;
    factor->exp   = NULL;
    factor->num   = 0;
    factor->alloc = 0;
}
